import 'dart:io';
import 'dart:async';
import 'dart:typed_data';
import 'dart:convert';
import 'package:intl/intl.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart' show rootBundle;
import 'package:dio/dio.dart';
import 'package:cookie_jar/cookie_jar.dart';
import 'package:path_provider/path_provider.dart';

// HTTP call type
enum apiCallType { GET, POST }

/// This class is used for creating a Dio object to be used for all HTTPS communication
/// to a specific endpoint.
class CdcrConnector {
  /// This class is used for creating a Dio object to be used for all HTTPS communication
  /// to a specific endpoint.
  ///
  /// __Required parameters__
  ///   ```
  /// [baseUri] represents the combination of the host name and the "base" service path.
  ///   This should not include the service endpojnt name or any params.
  /// [certFile] is the relative (to project) path of the certificate filename.
  /// [certPassword] is the password/passphrase enbedded in the certificate file.
  /// [apikey] is the application-specific API key...typically a UUID.
  /// [username] the username of the end-user...typically an AD username.
  /// [password] the password associated with the [username].
  ///   ```
  /// __Optional parameters__
  ///   ```
  /// [connectTimeout] is the timeout in milliseconds for sending data.
  ///   [Dio] will throw the [DioError] with [DioErrorType.SEND_TIMEOUT] type
  ///   when time out. The default is [15000] milliseconds.
  /// [receiveTimeout] is the timeout in milliseconds for receiving data.
  ///   [Dio] will throw the [DioError] with [DioErrorType.RECEIVE_TIMEOUT] type
  ///   when time out. The default is [15000] milliseconds.
  /// [followRedirects] set this property to [true] if this request should automatically
  ///   follow redirects. The default is [true].
  /// [maxRedirects] maximum number of redirects to follow when [followRedirects] is `true`.
  ///   If this number is exceeded an error event will be generated by the Dio package
  ///    The default is [6].
  /// [ignoreExpiredCookies] determines how to handle expired cookies.
  ///   When [false], cookies that have expired are NOT loaded or saved.
  ///   When [true], cookies that have expired are loaded and saved.
  /// [authenticateUser] authenticate the user or not. Typically not required on most
  ///   'get' services. Default is false/not.
  /// [loggingActivated] activate or deactive console logging. Default is deactivated.
  ///   ```
  CdcrConnector(Uri baseUri, String certFile, String certPassword,
      String apikey, String username, String password,
      {int connectTimeout = 15000,
      int receiveTimeout = 15000,
      bool followRedirects = true,
      int maxRedirects = 6,
      bool ignoreExpiredCookies = false,
      bool authenticateUser = false,
      bool loggingActivated = false}) {
    // Required params
    _baseUri = baseUri;
    _privateKeyFile = certFile;
    _certPassword = certPassword;
    _apikey = apikey;
    _username = username;
    _password = password;
    // Optional params
    _connectTimeout = connectTimeout;
    _receiveTimeout = receiveTimeout;
    _followRedirects = followRedirects;
    _maxRedirects = maxRedirects;
    _ignoreExpiredCookies = ignoreExpiredCookies;
    _loggingActivated = loggingActivated;
    // Update private data
    _headers['username'] = _username;
    _headers['password'] = _password;
    _options.baseUrl = _baseUrl = _baseUri.toString();
    _options.connectTimeout = _connectTimeout;
    _options.receiveTimeout = _receiveTimeout;
    _options.followRedirects = _followRedirects;
    _options.maxRedirects = _maxRedirects;
    _userAuthenticated = authenticateUser ? false : _userAuthenticated;
  }

  // Private declarations
  static Uri _baseUri;
  static String _baseUrl;
  static String _certPassword;
  static String _privateKeyFile;
  static String _apikey;
  static String _username;
  static String _password;
  static int _connectTimeout = 15000;
  static int _receiveTimeout = 15000;
  static int _maxRedirects = 6;
  static bool _followRedirects = true;
  static bool _ignoreExpiredCookies = false;
  static bool _loggingActivated = false;
  static bool _userAuthenticated = false;

  // Communications options used in DIO instance construction
  static BaseOptions _options = new BaseOptions(
    responseType: ResponseType.plain,
  );

  // Add the header attributes to send in the request
  static var _headers = {
    'username': '',
    'password': '',
  };

  // Error map
  static var _errorInfo = {
    'status': '',
    'errCode': '',
    'errDesc': '',
  };

  bool get isUserAuthenticated => _userAuthenticated == true;
  String get httpStatus => _errorInfo['status'];
  String get svrErrCode => _errorInfo['errCode'];
  String get svrErrDesc => _errorInfo['errDesc'];

  //
  // Print to console log
  void _printLog(String text, {bool printTS = true}) {
    if (_loggingActivated) {
      String now = DateFormat("MM-dd HH:mm:ss.SSS")
          .format(new DateTime.now())
          .toString();
      if (printTS)
        debugPrint("$now - $text", wrapWidth: 1024);
      else
        debugPrint(text, wrapWidth: 1024);
    }
  }

  //
  // Print to console log
  void _setErrInfo(String status, String errCode, String errDesc) {
    _errorInfo['status'] = status;
    _errorInfo['errCode'] = errCode;
    _errorInfo['errDesc'] = errDesc;
  }

  //
  // Create an error reply in JSON format
  String _createJsonError(String errMsg) {
    Map<String, dynamic> errResponse = {
      "status": "ERROR",
      "data": {"staffId": "NA", "token": ""},
      "message": errMsg
    };
    return json.encode(errResponse);
  }

  //
  /// Public method used for deleting cookies when a cookie/token has expired.
  void deleteCookies() async {
    // Set the location to load/store cookies.
    Directory appDocDir = await getApplicationDocumentsDirectory();
    PersistCookieJar cookieJar = new PersistCookieJar(
      dir: "${appDocDir.path}/cookies",
      ignoreExpires: _ignoreExpiredCookies,
    );
    // Delete all cookies for the desired base URI.
    cookieJar.delete(_baseUri, true);
    _printLog("Deleted cookies for $_baseUrl");
  }

  //
  /// Public method to force user authentication on next API call
  void resetUser() {
    _userAuthenticated = false;
  }

  //
  /// Public method used for calling a specific API/Service.
  ///   ```
  /// [callType] is either a GET or a POST.
  /// [serviceName] should be prefixed with a '/'. Path not required as it
  ///   was specified in the class construction.
  /// [data] is optional and is used only on POST calls.
  /// [sendCredOnGet] determines whether or not to send user credentials on GET actions.
  ///   Default is false.
  ///   ```
  Future<String> callAPI(apiCallType callType, String serviceName,
      {Map<String, String> data, bool sendCredOnGet = false}) async {
    //
    // Create the DIO instance
    Dio _dio = new Dio(_options);

    // Set the location to load/store cookies.
    Directory _appDocDir = await getApplicationDocumentsDirectory();
    PersistCookieJar _cookieJar = new PersistCookieJar(
      dir: "${_appDocDir.path}/cookies",
      ignoreExpires: _ignoreExpiredCookies,
    );

    // Each Dio instance has an interceptor by which you can intercept requests or
    // responses before they are handled by then or catchError. Contains a
    // [RequestInterceptor] and a [ResponseInterceptor] instance.
    _dio.interceptors..add(CookieManager(_cookieJar));
    if (_loggingActivated) _dio.interceptors.add(LogInterceptor());

    // Load the client certificate
    // * https://stackoverflow.com/questions/52355264/client-certificate-authentication-in-flutter-dart-languege/52359591
    Uint8List privateKey =
        (await rootBundle.load(_privateKeyFile)).buffer.asUint8List();
    SecurityContext sc = new SecurityContext();
    sc.useCertificateChainBytes(privateKey, password: _certPassword);
    sc.usePrivateKeyBytes(privateKey, password: _certPassword);
    (_dio.httpClientAdapter as DefaultHttpClientAdapter).onHttpClientCreate =
        (client) {
      HttpClient httpClient = new HttpClient(context: sc);
      httpClient.badCertificateCallback =
          (X509Certificate cert, String host, int port) => true;
      return httpClient;
    };

    // If the user hasn't authenticated yet, then authenticate them.
    if (!_userAuthenticated) {
      // Always delete old cookies before authenticating.
      _cookieJar.delete(_baseUri, true);
      _printLog("Deleted cookies for $_baseUrl");
      // Reset user credentials.
      _headers['username'] = _username;
      _headers['password'] = _password;
      // Authenticat the user.
      _userAuthenticated = await _authenticateUser(_dio);
    }
    // Send the requested API for an authenticated user. Otherwise, return null.
    return (_userAuthenticated)
        ? await _sendAPI(_dio, callType, serviceName,
            data: data, sendCredOnGet: sendCredOnGet)
        : null;
  } // end of callAPI()

  //
  // Use this method to strictly authenticate the user. Not intended for sending to the endpoint.
  Future<bool> _authenticateUser(Dio dio) async {
    int opCnt = 0, bufLen = 0, dataLen = 0;
    bool isError = false;
    Response dioResponse;
    // Send the authentication request
    _printLog(
        '****************************************************************',
        printTS: false);
    _printLog("******** BEGIN - AUTH REQUEST ********");
    try {
      isError = false;
      opCnt = bufLen = 0;

      dioResponse = await dio.get(
        _baseUrl,
        onReceiveProgress: (int receive, int total) {
          bufLen = receive;
          ++opCnt;
          _printLog(
              "--------------->Receive$opCnt: $receive bytes out of $total");
        },
        options: Options(
          headers: _headers,
          contentType: ContentType("text", "plain", charset: "utf-8"),
        ),
      );
    } on DioError catch (e) {
      // The request was made and the server responded with a status code
      // that falls out of the range of 2xx and is also not a 404 (Not Found).
      if (e.response != null) {
        if (e.response.statusCode != 404) {
          isError = true;
          _setErrInfo(e.response.statusCode.toString(), "0", e.message);
          _printLog("******** AUTH ERROR Start ********");
          _printLog("HTTP StatusCode: " + e.response.statusCode.toString(),
              printTS: false);
          _printLog("Error Type: " + e.type.toString(), printTS: false);
          _printLog("Error Message: " + e.message, printTS: false);
          _printLog("Request Headers: " + e.response.request.headers.toString(),
              printTS: false);
          _printLog("Reply Headers: " + e.response.headers.toString(),
              printTS: false);
          dataLen = e.response.data.toString().length;
          if (dataLen != 0)
            _printLog(
                "Reply Data: " +
                    e.response.data.toString().substring(0, dataLen),
                printTS: false);
          _printLog("******** AUTH ERROR End   ********");
        }
      } else {
        // The request err'd before reaching the service endpoint.
        isError = true;
        _setErrInfo("500", "0", "Authentication Error:\n" + e.message);
        _printLog("Error Type: " + e.type.toString(), printTS: false);
        _printLog("Error Message: " + e.message, printTS: false);
        if (e.message.toString() != e.error.toString())
          _printLog("Error Object: " + e.error.toString(), printTS: false);
      }
    } catch (er) {
      // An unexpected error occurrred during setup or send of the request.
      isError = true;
      _setErrInfo("500", "0", "Unknown AUTH Error!" + er);
      _printLog("******** AUTH ERROR Start ********");
      _printLog("Unknown AUTH Error: " + er, printTS: false);
      _printLog("******** AUTH ERROR End   ********");
    }

    if (!isError && dioResponse != null && dioResponse?.statusCode != 404) {
      isError = true;
      _setErrInfo("403", "0",
          "Authentication request failed!\nInvalid username or password.");
      _printLog("******** AUTH ERROR Start ********");
      _printLog('HTTP StatusCode: ${dioResponse.statusCode}', printTS: false);
      _printLog(
          'Authentication request failed!\nInvalid username or password. Please correct & try again.',
          printTS: false);
      int i = 0;
      dioResponse.redirects.forEach((rd) {
        ++i;
        _printLog('Redirect Code $i: ${rd.statusCode}', printTS: false);
        _printLog('Redirect Path $i: ${rd.location}', printTS: false);
      });
      //_printLog('Reply Data: \n' + dioResponse.data.toString(), printTS: false);
      _printLog("******** AUTH ERROR End   ********");
    }

    _printLog("******** END   - AUTH REQUEST ********");
    if (isError) {
      _printLog("\n", printTS: false); // Force print buffer to flush
      return false;
    }

    if (dioResponse != null && bufLen > 0) {
      _printLog(
          '========================= BEGIN - AUTH REPLY DATA ==========================');
      dataLen = dioResponse.data.toString().length;
      _printLog(dioResponse.data.toString().substring(0, dataLen),
          printTS: false);
      _printLog(
          '========================= END   - AUTH REPLY DATA ==========================');
      _printLog("\n", printTS: false); // Force print buffer to flush
    }
    return true;
  }

  //
  // Use this method to strictly send an API to the endpoint.
  Future<String> _sendAPI(Dio dio, apiCallType callType, String serviceName,
      {Map<String, String> data, bool sendCredOnGet = false}) async {
    bool isError = false;
    int opCnt = 0, bufLen = 0, dataLen = 0;
    Response dioResponse;
    String strResponse;

    // Send the service request
    _printLog("******** BEGIN - API REQUEST ********");
    try {
      isError = false;
      opCnt = bufLen = 0;

      switch (callType) {
        case apiCallType.GET:
          // Create the query params
          Map<String, String> qryPrms = {"shs": _apikey};
          if (data != null) qryPrms.addAll(data);
          // Send a GET
          dioResponse = await dio.get(
            serviceName,
            onReceiveProgress: (int receive, int total) {
              bufLen = receive;
              ++opCnt;
              _printLog(
                  "--------------->Receive$opCnt: $receive bytes out of $total");
            },
            queryParameters: qryPrms,
            options: Options(
              headers: (sendCredOnGet) ? _headers : null,
              contentType: ContentType("application", "json", charset: "utf-8"),
            ),
          );
          break;
        case apiCallType.POST:
          // Send a POST
          dioResponse = await dio.post(
            serviceName,
            data: data,
            onSendProgress: (int sent, int total) {
              bufLen = sent;
              ++opCnt;
              _printLog(
                  "--------------->Send$opCnt: $sent bytes out of $total");
            },
            queryParameters: {"shs": _apikey},
            options: Options(
              headers: _headers,
              contentType: ContentType("application", "json", charset: "utf-8"),
            ),
          );
          break;
        default:
      }
    } on DioError catch (e) {
      isError = true;
      // The request was made and the server responded with a status code
      // that falls out of the range of 2xx and is also not 304.
      _printLog("******** API ERROR Start ********");
      if (e.response != null) {
        _setErrInfo(e.response.statusCode.toString(), "0", e.message);
        _printLog("HTTP StatusCode: " + e.response.statusCode.toString(),
            printTS: false);
        _printLog("Error Type: " + e.type.toString(), printTS: false);
        _printLog("Error Message: " + e.message, printTS: false);
        _printLog("Request Headers: " + e.response.request.headers.toString(),
            printTS: false);
        _printLog("Reply Headers: " + e.response.headers.toString(),
            printTS: false);
        dataLen = e.response.data.toString().length;
        if (dataLen != 0)
          _printLog(
              "Reply Data: " + e.response.data.toString().substring(0, dataLen),
              printTS: false);
      } else {
        // The request err'd before reaching the service endpoint.
        _setErrInfo("500", "0", "API Error:\n" + e.message);
        _printLog("Error Type: " + e.type.toString(), printTS: false);
        _printLog("Error Message: " + e.message, printTS: false);
        if (e.message.toString() != e.error.toString())
          _printLog("Error Object: " + e.error.toString(), printTS: false);
      }
      _printLog("******** API ERROR End   ********");
    } catch (err) {
      // An unexpected error occurrred during setup or send of the request.
      isError = true;
      _setErrInfo("500", "0", "Unknown API Error!: " + err);
      _printLog("******** API ERROR Start ********");
      _printLog("Unknown API Error: " + err, printTS: false);
      _printLog("******** API ERROR End   ********");
    }
    _printLog("******** END   - API REQUEST ********");
    if (isError) {
      _printLog("\n", printTS: false); // Force print buffer to flush
      return strResponse;
    }

    // Check for empty payload and generate an error if empty
    if (dioResponse == null || bufLen < 1) {
      dioResponse.data = _createJsonError(
          "Got a successful API response, but payload is empty.");
      bufLen = dioResponse.data.toString().length;
    } else
    // Check for HTML instead of JSON payload
    if (dioResponse.data.toString().toLowerCase().contains("<html>")) {
      bool invalidCredentials = false;
      dioResponse.redirects.forEach((rd) {
        if (rd.statusCode == 302 &&
                (rd.location.toString().toLowerCase().contains("logout.php")) ||
            (rd.location.toString().toLowerCase().contains("hangup.php"))) {
          invalidCredentials = true;
        }
      });
      if (invalidCredentials)
        dioResponse.data = _createJsonError("Invalid username or password!");
      else
        dioResponse.data = _createJsonError(
            "Malformed response message in HTML rather than JSON.");
      bufLen = dioResponse.data.toString().length;
    }

    _printLog(
        '========================= BEGIN - API REPLY DATA ==========================');
    Map parsedJson = json.decode(dioResponse.data.toString());
    if (parsedJson['status'].toString() != 'OK') {
      _setErrInfo(
          dioResponse.statusCode.toString(),
          parsedJson['status'].toString(),
          "API Error:\n" + parsedJson['message'].toString());
      _printLog("******** API ERROR Start ********");
      _printLog(dioResponse.data.toString(), printTS: false);
      _printLog("******** API ERROR End   ********");
    } else {
      _setErrInfo(dioResponse.statusCode.toString(), '', '');
      _printLog(
          'Service (' +
              serviceName.substring(1) +
              ') Reply Length: ' +
              bufLen.toString(),
          printTS: false);
      if (bufLen > 10000) {
        dataLen = dioResponse.data.toString().length;
        _printLog('Too much data to log!', printTS: false);
        _printLog(
            '========================= END   - API REPLY DATA ==========================');
        _printLog("\n", printTS: false); // Force print buffer to flush
      } else {
        dataLen = dioResponse.data.toString().length;
        _printLog(dioResponse.data.toString().substring(0, dataLen),
            printTS: false);
      }
      strResponse = dioResponse.data.toString().substring(0, dataLen);
    }
    _printLog(
        '========================= END   - API REPLY DATA ==========================');
    _printLog("\n", printTS: false); // Force print buffer to flush
    return strResponse;
  } //  end of sendAPI()

} // end of class
